import copy
from dataclasses import dataclass

from ruamel.yaml import CommentedMap

from exasol.toolbox.util.workflows import logger
from exasol.toolbox.util.workflows.exceptions import (
    YamlJobValueError,
    YamlStepIdValueError,
)
from exasol.toolbox.util.workflows.patch_workflow import (
    ActionType,
    WorkflowCommentedMap,
)
from exasol.toolbox.util.workflows.render_yaml import YamlRenderer


@dataclass(frozen=True)
class WorkflowRenderer(YamlRenderer):
    """
    The :class:`WorkflowRenderer` renders a workflow template provided by the PTB into
    a final workflow. It renders the final workflow by:
      - resolving Jinja variables.
      - standardizing formatting via ruamel.yaml for a consistent output.
    """

    patch_yaml: WorkflowCommentedMap | None

    def render(self) -> str:
        """
        Render the template to the contents of a valid GitHub workflow.
        """
        workflow_dict = self.get_yaml_dict()

        if self.patch_yaml:
            logger.debug("Customize workflow with `patch_yaml`")
            workflow_modifier = WorkflowModifier(
                workflow_dict=workflow_dict, patch_yaml=self.patch_yaml
            )
            workflow_dict = workflow_modifier.get_patched_workflow()

        return self.get_as_string(workflow_dict)


@dataclass
class WorkflowModifier:
    workflow_dict: CommentedMap
    patch_yaml: WorkflowCommentedMap

    def __post_init__(self):
        # Perform deepcopy to ensure this instance owns its data
        self.workflow_dict = copy.deepcopy(self.workflow_dict)

    @property
    def jobs_dict(self) -> CommentedMap:
        return self.workflow_dict["jobs"]

    def _get_step_list(self, job_name: str) -> CommentedMap:
        self._verify_job_exists(job_name=job_name)
        return self.jobs_dict[job_name]["steps"]

    def _customize_steps(self, step_customizations: CommentedMap) -> None:
        """
        Customize the steps of jobs specified in `step_customizations` in a workflow
        (`workflow_dict`). If a `step_id` or its parent `job` cannot be found, an
        exception is raised.
        """
        for patch in step_customizations:
            job_name = patch["job"]
            patch_id = patch["step_id"]
            idx = self._get_step_index(job_name=job_name, step_id=patch_id)

            step_list = self._get_step_list(job_name=job_name)
            if patch["action"] == ActionType.REPLACE.value:
                logger.debug(
                    f"Replace YAML at step_id '{patch_id}' in job '{job_name}'"
                )
                step_list[idx : idx + 1] = patch["content"]

            elif patch["action"] == ActionType.INSERT_AFTER.value:
                logger.debug(
                    f"Insert YAML after step_id '{patch_id}' in job '{job_name}'"
                )
                step_list[idx + 1 : idx + 1] = patch["content"]

    def _get_step_index(self, job_name: str, step_id: str) -> int:
        steps = self._get_step_list(job_name=job_name)
        for index, step in enumerate(steps):
            if step["id"] == step_id:
                return index
        raise YamlStepIdValueError(step_id=step_id, job_name=job_name)

    def _remove_jobs(self, remove_jobs: CommentedMap) -> None:
        """
        Remove the jobs specified in `remove_jobs` in a workflow yaml (`workflow_dict`).
        If a `job` cannot be found, an exception is raised.
        """
        for job_name in remove_jobs:
            self._verify_job_exists(job_name)
            logger.debug(f"Remove job '{job_name}'")
            self.jobs_dict.pop(job_name)

    def _verify_job_exists(self, job_name: str) -> None:
        if job_name not in self.jobs_dict:
            raise YamlJobValueError(job_name=job_name)

    def get_patched_workflow(self):
        """
        Patch the `workflow_dict`. As dictionaries are mutable structures, we directly
        take advantage of this by having it modified in this class's internal methods
        without explicit returns.
        """

        if remove_jobs := self.patch_yaml.get("remove_jobs", {}):
            self._remove_jobs(remove_jobs=remove_jobs)

        if step_customizations := self.patch_yaml.get("step_customizations", {}):
            self._customize_steps(step_customizations=step_customizations)

        return self.workflow_dict



import io
from dataclasses import dataclass
from inspect import cleandoc
from pathlib import Path
from typing import Any

from jinja2 import (
    Environment,
    StrictUndefined,
    TemplateError,
)
from ruamel.yaml import (
    YAML,
    CommentedMap,
)
from ruamel.yaml.error import YAMLError

from exasol.toolbox.util.workflows import logger
from exasol.toolbox.util.workflows.exceptions import (
    TemplateRenderingError,
    YamlOutputError,
    YamlParsingError,
)

jinja_env = Environment(
    variable_start_string="((",
    variable_end_string="))",
    autoescape=True,
    # This requires that all Jinja variables must be defined in the provided
    # dictionary. If not, then a `jinja2.exceptions.UndefinedError` exception
    # will be raised.
    undefined=StrictUndefined,
)


@dataclass(frozen=True)
class YamlRenderer:
    """
    The :class:`YamlRenderer` provides a standardised interface for rendering YAML
    files within the PTB. To simplify configuration and reduce manual coordination,
    use Jinja variables as defined in :meth:`BaseConfig.github_template_dict` in your
    YAML files.
    """

    github_template_dict: dict[str, Any]
    file_path: Path

    @staticmethod
    def _get_standard_yaml() -> YAML:
        """
        Prepare standard YAML class.
        """
        yaml = YAML()
        yaml.width = 200
        yaml.preserve_quotes = True
        yaml.sort_base_mapping_type_on_output = False  # type: ignore
        yaml.indent(mapping=2, sequence=4, offset=2)
        return yaml

    def _render_with_jinja(self, input_str: str) -> str:
        """
        Render the template with Jinja.
        """
        logger.debug(
            "Render template with Jinja",
            jinja_dict_source="PROJECT_CONFIG.github_template_dict",
            jinja_dict_values=self.github_template_dict,
        )
        jinja_template = jinja_env.from_string(input_str)
        return jinja_template.render(self.github_template_dict)

    def get_yaml_dict(self) -> CommentedMap:
        """
        Load a file as a CommentedMap (dictionary form of a YAML), after
        rendering it with Jinja.
        """
        raw_content = self.file_path.read_text()
        try:
            workflow_string = self._render_with_jinja(raw_content)
            yaml = self._get_standard_yaml()
            logger.debug("Parse template with ruamel-yaml")
            return yaml.load(workflow_string)
        except TemplateError as ex:
            raise TemplateRenderingError(file_path=self.file_path) from ex
        except YAMLError as ex:
            raise YamlParsingError(file_path=self.file_path) from ex

    def get_as_string(self, yaml_dict: CommentedMap) -> str:
        """
        Output a YAML string.
        """
        yaml = self._get_standard_yaml()
        try:
            logger.debug("Output workflow as string")
            with io.StringIO() as stream:
                yaml.dump(yaml_dict, stream)
                workflow_string = stream.getvalue()
            return cleandoc(workflow_string)
        except YAMLError as ex:
            raise YamlOutputError(file_path=self.file_path) from ex



from collections.abc import Mapping
from pathlib import Path

import importlib_resources as resources

WORKFLOW_TEMPLATES_DIRECTORY = "exasol.toolbox.templates.github.workflows"


def get_workflow_templates() -> Mapping[str, Path]:
    """
    Returns a mapping for workflow templates, where the keys are filenames without the
    '.yml' extension and the values are the filepaths.
    """
    package_resources = resources.files(WORKFLOW_TEMPLATES_DIRECTORY)
    return {
        workflow_path.name.removesuffix(".yml"): Path(str(workflow_path))
        for workflow_path in package_resources.iterdir()
        if workflow_path.is_file() and workflow_path.name.endswith(".yml")
    }


WORKFLOW_TEMPLATE_OPTIONS = get_workflow_templates()



import json
import logging

import nox
from nox import Session

from exasol.toolbox.config import BaseConfig
from noxconfig import (
    PROJECT_CONFIG,
)

_log = logging.getLogger(__name__)


def _python_matrix(config: BaseConfig):
    return {"python-version": config.python_versions}


def _exasol_matrix(config: BaseConfig):
    return {"exasol-version": config.exasol_versions}


@nox.session(name="matrix:python", python=False)
def python_matrix(session: Session) -> None:
    """Output the build matrix for Python versions as JSON."""
    print(json.dumps(_python_matrix(PROJECT_CONFIG)))


@nox.session(name="matrix:exasol", python=False)
def exasol_matrix(session: Session) -> None:
    """Output the build matrix for Exasol versions as JSON."""
    print(json.dumps(_exasol_matrix(PROJECT_CONFIG)))


@nox.session(name="matrix:all", python=False)
def full_matrix(session: Session) -> None:
    """Output the full build matrix for Python & Exasol versions as JSON."""
    matrix = _python_matrix(PROJECT_CONFIG)
    matrix.update(_exasol_matrix(PROJECT_CONFIG))
    print(json.dumps(matrix))



from collections.abc import Mapping
from pathlib import Path


class YamlError(Exception):
    """
    Base exception for YAML errors.
    """

    message_template = "An error occurred with file: {file_path}"

    def __init__(self, file_path: Path, **kwargs):
        self.file_path = file_path
        # Format the template defined in the subclass
        message = self.message_template.format(file_path=file_path, **kwargs)
        super().__init__(message)


class YamlOutputError(YamlError):
    """
    Raised when the final workflow cannot be exported as a YAML file.
    This would likely indicate that one of the preceding transformation steps
    led to a format that is no longer able to be exported as a YAML file.
    """

    message_template = "File {file_path} could not be output by ruamel-yaml."


class YamlParsingError(YamlError):
    """
    Raised when the rendered template is not a valid YAML file, as it cannot be
    parsed by ruamel-yaml.
    """

    message_template = (
        "File {file_path} could not be parsed. Check for invalid YAML syntax."
    )


class TemplateRenderingError(YamlError):
    """
    Raised when Jinja fails to modify the template. It may be that a Jinja
    variable was not defined, a brace was not closed, etc.
    """

    message_template = (
        "File {file_path} failed to render. Check for Jinja-related errors."
    )


class InvalidWorkflowPatcherYamlError(YamlError):
    """
    Raised when the :class:`WorkflowPatcher` failed the validation constraints of
    :class:`WorkflowPatcherConfig`.
    """

    message_template = "File '{file_path}' is malformed; it failed Pydantic validation."


class InvalidWorkflowPatcherEntryError(YamlError):
    """
    Raised when the :class:`WorkflowPatcher` is used but one of the specified keys it
    listed does not exist in the relevant workflow template file.
    """

    message_template = (
        "In file '{file_path}', an entry '{entry}' does not exist in "
        "the workflow template. Please fix the entry."
    )


class YamlKeyError(Exception):
    """
    Base exception for when a specified value cannot be found in a YAML.
    """

    message_template = "An error occurred with job: '{job_name}'"

    def __init__(self, **kwargs):
        # Store all attributes dynamically (job_name, step_id, etc.)
        for key, value in kwargs.items():
            setattr(self, key, value)

        self._data = kwargs
        # Format the template using the passed-in arguments
        super().__init__(self.message_template.format(**kwargs))

    @property
    def entry(self) -> Mapping[str, str]:
        return self._data


class YamlJobValueError(YamlKeyError):
    """
    Raised when a job cannot be found in a YAML file.
    """

    message_template = "Job '{job_name}' could not be found"


class YamlStepIdValueError(YamlKeyError):
    """
    Raised when a step_id associated with a specific job cannot be found in a YAML file.
    """

    message_template = "Step_id '{step_id}' not found in job '{job_name}'"



from json import (
    dumps,
    loads,
)
from pathlib import Path

from exasol.toolbox.util.version import Version

PROJECT_ROOT = Path(__file__).resolve().parents[4]
COOKIECUTTER_JSON = PROJECT_ROOT / "project-template" / "cookiecutter.json"


def update_cookiecutter_default(version: Version) -> None:
    contents = COOKIECUTTER_JSON.read_text()
    contents_as_dict = loads(contents)

    contents_as_dict["exasol_toolbox_version_range"] = f">={version},<{version.major+1}"

    updated_contents = dumps(contents_as_dict, indent=2)
    COOKIECUTTER_JSON.write_text(updated_contents)



from __future__ import annotations

import nox
from nox import Session

from noxconfig import PROJECT_CONFIG


@nox.session(name="package:check", python=False)
def package_check(session: Session) -> None:
    """Checks whether your distribution’s long description will render correctly on PyPI

    This has more robust checks for rst documentation than markdown.
    """
    session.run("poetry", "build", "--project", PROJECT_CONFIG.root_path)
    session.run("twine", "check", PROJECT_CONFIG.root_path / "./dist/*")



import inspect
import warnings
from collections.abc import Callable
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Literal,
)

from packaging import version
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_validator,
)
from pydantic_core.core_schema import ValidationInfo

from exasol.toolbox.nox.plugin import (
    METHODS_SPECIFIED_FOR_HOOKS,
    PLUGIN_ATTR_NAME,
)
from exasol.toolbox.util.version import Version


def get_methods_with_hook_implementation(
    plugin_class: type[Any],
) -> tuple[tuple[str, Callable], ...]:
    """
    Get all methods from a plugin_class which were specified with a @hookimpl.
    """
    return tuple(
        (name, method)
        for name, method in inspect.getmembers(plugin_class, inspect.isroutine)
        if hasattr(method, PLUGIN_ATTR_NAME)
    )


def filter_not_specified_methods(
    methods: tuple[tuple[str, Callable], ...],
) -> tuple[str, ...]:
    """
    Filter methods which were specified with a @hookimpl but where not specified
    in ``exasol.toolbox.nox.plugins.NoxTasks``.
    """
    return tuple(name for name, _ in methods if name not in METHODS_SPECIFIED_FOR_HOOKS)


def validate_plugin_hook(plugin_class: type[Any]):
    """
    Validate methods in a class for at least one pluggy @hookimpl marker and verifies
    that this method is also specified in ``exasol.toolbox.nox.plugins.NoxTasks``.
    """
    methods_with_hook = get_methods_with_hook_implementation(plugin_class=plugin_class)

    if len(methods_with_hook) == 0:
        raise ValueError(
            f"No methods in `{plugin_class.__name__}` were found to be decorated"
            "with `@hookimpl`. The `@hookimpl` decorator indicates that this"
            "will be used with pluggy and used in specific nox sessions."
            "Without it, this class does not modify any nox sessions."
        )

    if not_specified_methods := filter_not_specified_methods(methods_with_hook):
        raise ValueError(
            f"{len(not_specified_methods)} method(s) were "
            "decorated with `@hookimpl`, but these methods were not "
            "specified in `exasol.toolbox.nox.plugins.NoxTasks`: "
            f"{not_specified_methods}. The `@hookimpl` decorator indicates "
            "that these methods will be used by pluggy to modify specific nox sessions."
            "If the method was not previously specified, then no nox sessions will"
            "be modified. The `@hookimpl` is only used by nox sessions provided by the"
            "pyexasol-toolbox and not ones created for just your project."
        )

    return plugin_class


def valid_version_string(version_string: str) -> str:
    Version.from_string(version_string)
    return version_string


ValidPluginHook = Annotated[type[Any], AfterValidator(validate_plugin_hook)]
ValidVersionStr = Annotated[str, AfterValidator(valid_version_string)]

DEFAULT_EXCLUDED_PATHS = {
    ".eggs",
    ".html-documentation",
    ".poetry",
    ".sonar",
    ".venv",
    "dist",
    "venv",
}


class DependencyManager(BaseModel):
    # Restricted to only allow "poetry" at the moment
    name: Literal["poetry"]
    version: ValidVersionStr

    @field_validator("version")
    @classmethod
    def check_minimum_version(cls, v: str, info: ValidationInfo) -> str:
        tool = info.data.get("name")
        if tool == "poetry":
            prefix = "Poetry version "
            if version.parse(v) < (min_version := version.parse("2.1.4")):
                raise ValueError(prefix + f"must be >= {min_version}")
            elif version.parse(v) >= (max_version := version.parse("3.0.0")):
                raise ValueError(prefix + f"must be < {max_version}")
            elif version.parse(v) > (current_version := version.parse("2.3.0")):
                warnings.warn(prefix + f"exceeds last tested version {current_version}")
        return v


class BaseConfig(BaseModel):
    """
    Basic configuration for projects using the PTB

    This configuration class defines the necessary attributes for using the PTB's
    various nox sessions and GitHub CI workflows. Defaults are provided via the
    attributes, which makes it easy for a specific project to modify these values
    as needed. There are properties provided which automate some configuration aspects
    so that the project-specific modifications are reduced.

    pydantic has been used so that altered attributes are quickly validated.
    This allows for immediate feedback, instead of waiting for various failed CI
    runs.
    """

    project_name: str = Field(description="Name of the project")
    root_path: Path = Field(description="Root directory of the project")
    python_versions: tuple[ValidVersionStr, ...] = Field(
        default=("3.10", "3.11", "3.12", "3.13", "3.14"),
        description="Python versions to use in running CI workflows",
    )

    exasol_versions: tuple[ValidVersionStr, ...] = Field(
        default=("7.1.30", "8.29.13", "2025.1.8"),
        description="Exasol versions to use in running CI workflows for integration tests using the DB",
    )
    create_major_version_tags: bool = Field(
        default=False,
        description="If true, creates also the major version tags (v*) automatically",
    )
    add_to_excluded_python_paths: tuple[str, ...] = Field(
        default=(),
        description="""
        This is used to extend the default excluded_python_paths. If a more general
        path that would be seen in other projects, like .venv, needs to be added into
        this argument, please instead modify the
        ``exasol.toolbox.config.DEFAULT_EXCLUDED_PATHS``.
        """,
    )
    plugins_for_nox_sessions: tuple[ValidPluginHook, ...] = Field(
        default=(),
        description="""
        This is used to provide hooks to extend one or more of the Nox sessions provided
        by the python-toolbox. As described on the plugins pages:\n\n
        * https://exasol.github.io/python-toolbox/main/user_guide/customization.html#plugins
        * https://exasol.github.io/python-toolbox/main/developer_guide/plugins.html

        possible plugin options are defined in ``exasol.toolbox.nox.plugins.NoxTasks``.
        """,
    )
    dependency_manager: DependencyManager = Field(
        default=DependencyManager(name="poetry", version="2.3.0"),
        description="""
        This is used to define which dependency manager is used to install dependencies
        in the CI. For more details on which PTB version pairs with which
        dependency manager, see:\n\n
        * https://exasol.github.io/python-toolbox/main/user_guide/dependencies.html
        """,
    )
    os_version: str = Field(
        default="ubuntu-24.04",
        pattern=r"^ubuntu-.*",
        description="""
        This is used to set the OS-runner in the GitHub workflows that are
        provided as templates from the PTB. Currently, only ubuntu-based runners
        are supported.
        """,
    )
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    @computed_field  # type: ignore[misc]
    @property
    def documentation_path(self) -> Path:
        """
        Path where the documentation for the project is stored.
        """
        return self.root_path / "doc"

    @computed_field  # type: ignore[misc]
    @property
    def minimum_python_version(self) -> str:
        """
        Minimum Python version declared from the ``python_versions`` list

        This is used in specific testing scenarios where it would be either
        costly to run the tests for all ``python_versions`` or we need a single metric.
        """
        versioned = [Version.from_string(v) for v in self.python_versions]
        min_version = min(versioned)
        index_min_version = versioned.index(min_version)
        return self.python_versions[index_min_version]

    @computed_field  # type: ignore[misc]
    @property
    def excluded_python_paths(self) -> tuple[str, ...]:
        """
        There are certain nox sessions:
          - format:check
          - format:fix
          - lint:code
          - lint:security
          - lint:typing

        where it is desired to restrict which Python files are considered within the
        ``PROJECT_CONFIG.source_code_path`` path, like excluding ``dist``, ``.eggs``.
        As such, this property is used to exclude such undesired paths.
        """
        return tuple(
            sorted(DEFAULT_EXCLUDED_PATHS.union(set(self.add_to_excluded_python_paths)))
        )

    @computed_field  # type: ignore[misc]
    @property
    def pyupgrade_argument(self) -> tuple[str, ...]:
        """
        Default argument to :func:`exasol.toolbox._format._pyupgrade`.

        It uses the minimum Python version to ensure compatibility with all supported
        versions of a project.
        """
        version_parts = self.minimum_python_version.split(".")[:2]
        version_number = "".join(version_parts)
        return (f"--py{version_number}-plus",)

    @computed_field  # type: ignore[misc]
    @property
    def sonar_code_path(self) -> Path:
        """
        Relative path needed in Nox session ``sonar:check`` to create the coverage XML.
        """
        return self.source_code_path.relative_to(self.root_path)

    @computed_field  # type: ignore[misc]
    @property
    def source_code_path(self) -> Path:
        """
        Path to the source code of the project.
        """
        return self.root_path / "exasol" / self.project_name

    @computed_field  # type: ignore[misc]
    @property
    def version_filepath(self) -> Path:
        """
        Path to the ``version.py`` file included in the project. This is an
        autogenerated file which contains the version of the code. It is maintained by
        the nox sessions ``version:check`` and ``release:prepare``.
        """
        return self.source_code_path / "version.py"

    @computed_field  # type: ignore[misc]
    @property
    def github_workflow_directory(self) -> Path:
        """
        Path to the GitHub workflow directory.
        """
        return self.root_path / ".github" / "workflows"

    @computed_field  # type: ignore[misc]
    @property
    def github_template_dict(self) -> dict[str, Any]:
        """
        Dictionary of variables to dynamically render Jinja2 templates into valid YAML
        configurations.
        """
        return {
            "dependency_manager_version": self.dependency_manager.version,
            "minimum_python_version": self.minimum_python_version,
            "os_version": self.os_version,
            "python_versions": self.python_versions,
        }

    @computed_field  # type: ignore[misc]
    @property
    def github_workflow_patcher_yaml(self) -> Path | None:
        """
        For customizing the GitHub workflow templates provided by the PTB,
        a project can define a ``.workflow-patcher.yml`` file containing
        instructions to delete or modify jobs in the PTB template.
        Modification includes replacing and inserting steps.
        """
        workflow_patcher_yaml = self.root_path / ".workflow-patcher.yml"
        if workflow_patcher_yaml.exists():
            return workflow_patcher_yaml
        return None



from enum import Enum
from functools import cached_property
from typing import (
    Annotated,
    Any,
    TypeAlias,
)

from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
)
from ruamel.yaml import CommentedMap
from structlog.contextvars import bound_contextvars

from exasol.toolbox.util.workflows import logger
from exasol.toolbox.util.workflows.exceptions import InvalidWorkflowPatcherYamlError
from exasol.toolbox.util.workflows.render_yaml import YamlRenderer
from exasol.toolbox.util.workflows.templates import WORKFLOW_TEMPLATE_OPTIONS


class ActionType(str, Enum):
    INSERT_AFTER = "INSERT_AFTER"
    REPLACE = "REPLACE"


class StepContent(BaseModel):
    """
    The :class:`StepContent` is used to lightly validate the content which
    would be used to REPLACE or INSERT_AFTER the specified step in the GitHub workflow.

    With the value ``ConfigDict(extra="allow")``, this model allows for further fields
    (e.g. ``dummy``) to be specified without any validation. This design choice was
    intentional, as GitHub already allows additional fields and may specify more fields
    than what has been specified in this model.

    As the validation here is light, it is left to GitHub to validate the content.
    For further information on what is allowed & expected for the fields, refer to
    `GitHub's documentation on jobs.<job_id>.steps <https://docs.github.com/en/actions/reference/workflows-and-actions/workflow-syntax#jobsjob_idsteps>`__.
    """

    model_config = ConfigDict(extra="allow")  # This allows extra fields

    name: str
    id: str
    uses: str | None = None
    run: str | None = None
    with_: dict[str, Any] | None = Field(None, alias="with")
    env: dict[str, str] | None = None


class StepCustomization(BaseModel):
    """
    The :class:`StepCustomization` is used to specify the desired modification.
    An ``action`` of ``ActionType``:

    * ``REPLACE`` - means that the contents of the specified ``step_id`` should be
      replaced with whatever ``content`` is provided.
    * ``INSERT_AFTER`` - means that the specified `content` should be inserted after
      the specified ``step_id``.
    """

    action: ActionType
    job: str
    step_id: str
    content: list[StepContent]


def validate_workflow_name(workflow_name: str) -> str:
    """
    Validates that the given ``workflow_name`` is a valid workflow name provided by
    the PTB.
    """
    if workflow_name not in WORKFLOW_TEMPLATE_OPTIONS.keys():
        raise ValueError(
            f"Invalid workflow: {workflow_name}. Must be one of {WORKFLOW_TEMPLATE_OPTIONS.keys()}"
        )
    return workflow_name


WorkflowName = Annotated[str, AfterValidator(validate_workflow_name)]


class Workflow(BaseModel):
    """
    The :class:`Workflow` is used to specify which workflow should be modified.
    This is determined by the workflow ``name``. A workflow can be modified by
    specifying:

    * ``remove_jobs`` - job names in this list will be removed from the workflow.
    * ``step_customization`` - items in this list indicate which job's step
      should be modified.
    """

    name: WorkflowName
    remove_jobs: list[str] = Field(default_factory=list)
    step_customizations: list[StepCustomization] = Field(default_factory=list)


class WorkflowPatcherConfig(BaseModel):
    """
    The :class:`WorkflowPatcherConfig` is used to validate the expected format
    for the ``.workflow-patcher.yml``, which is used to modify the workflow
    templates provided by the PTB.
    """

    workflows: list[Workflow]


WorkflowCommentedMap: TypeAlias = Annotated[
    CommentedMap, f"This CommentedMap is structured according to `{Workflow.__name__}`"
]


class WorkflowPatcher(YamlRenderer):
    """
    The :class:`WorkflowPatcher` enables users to define a YAML file
    to customize PTB-provided workflows by removing or modifying jobs in the file.
    A job can be modified by replacing or inserting steps.
    The provided YAML file must meet the conditions of :class:`WorkflowPatcherConfig`.
    """

    @cached_property
    def content(self) -> CommentedMap:
        """
        The loaded YAML content. It loads on first access and stays cached even though
        the class is frozen.
        """
        with bound_contextvars(template_file_name=self.file_path.name):
            logger.info("Load workflow patcher: %s", self.file_path.name)
            loaded_yaml = self.get_yaml_dict()
            try:
                logger.debug("Validate workflow patcher with Pydantic")
                WorkflowPatcherConfig.model_validate(loaded_yaml)
                return loaded_yaml
            except ValidationError as ex:
                raise InvalidWorkflowPatcherYamlError(file_path=self.file_path) from ex

    def extract_by_workflow(self, workflow_name: str) -> WorkflowCommentedMap | None:
        """
        Extract from the `content` where `name` matches the `workflow_name`. If the
        workflow is not found, then `None` is returned. It is an expected and common
        use case that the `WorkflowPatcher` would only modify a few workflows and not
        all of them.
        """
        inner_content = self.content["workflows"]
        for workflow in inner_content:
            if workflow["name"] == workflow_name:
                return workflow
        return None



import logging
import os

import structlog
from structlog.dev import ConsoleRenderer
from structlog.processors import (
    CallsiteParameter,
    CallsiteParameterAdder,
    TimeStamper,
    add_log_level,
    format_exc_info,
)

log_level = os.getenv("LOG_LEVEL", "INFO").upper()

structlog.configure(
    wrapper_class=structlog.make_filtering_bound_logger(getattr(logging, log_level)),
    processors=[
        # 1. Enrich the data first
        add_log_level,
        TimeStamper(fmt="iso"),
        CallsiteParameterAdder(
            {
                CallsiteParameter.MODULE,
                CallsiteParameter.FUNC_NAME,
            }
        ),
        # 2. Handle exceptions
        format_exc_info,
        # 3. Rendering option
        ConsoleRenderer(),
    ],
)



from __future__ import annotations

import argparse

import nox
from nox import Session

from exasol.toolbox.util.workflows.workflow import (
    WORKFLOW_CHOICES,
    update_workflow,
)
from noxconfig import PROJECT_CONFIG


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="nox -s workflow:generate",
        usage="nox -s workflow:generate -- [-h]  <workflow_choice>",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "workflow_choice",
        choices=WORKFLOW_CHOICES,
        help="Select one workflow or 'all' to all workflows.",
    )
    return parser


@nox.session(name="workflow:generate", python=False)
def generate_workflow(session: Session) -> None:
    """
    Generate or update the specified GitHub workflow or all of them.
    """
    parser = _create_parser()
    args = parser.parse_args(session.posargs)

    PROJECT_CONFIG.github_workflow_directory.mkdir(parents=True, exist_ok=True)

    update_workflow(workflow_choice=args.workflow_choice, config=PROJECT_CONFIG)



from collections.abc import Mapping
from pathlib import Path
from typing import (
    Annotated,
    Any,
    Final,
)

from pydantic import (
    BaseModel,
    ConfigDict,
)
from structlog.contextvars import (
    bound_contextvars,
)

from exasol.toolbox.config import BaseConfig
from exasol.toolbox.util.workflows import logger
from exasol.toolbox.util.workflows.exceptions import (
    InvalidWorkflowPatcherEntryError,
    YamlError,
    YamlKeyError,
)
from exasol.toolbox.util.workflows.patch_workflow import (
    WorkflowCommentedMap,
    WorkflowPatcher,
)
from exasol.toolbox.util.workflows.process_template import WorkflowRenderer
from exasol.toolbox.util.workflows.templates import WORKFLOW_TEMPLATE_OPTIONS

ALL: Final[str] = "all"
WORKFLOW_CHOICES: Final[list[str]] = [ALL, *WORKFLOW_TEMPLATE_OPTIONS.keys()]

WorkflowChoice = Annotated[str, f"Should be a value from {WORKFLOW_CHOICES}"]


class Workflow(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    content: str

    @classmethod
    def load_from_template(
        cls,
        file_path: Path,
        github_template_dict: dict[str, Any],
        patch_yaml: WorkflowCommentedMap | None = None,
    ):
        with bound_contextvars(template_file_name=file_path.name):
            logger.info("Load workflow template: %s", file_path.name)

            if not file_path.exists():
                raise FileNotFoundError(file_path)

            try:
                workflow_renderer = WorkflowRenderer(
                    github_template_dict=github_template_dict,
                    file_path=file_path,
                    patch_yaml=patch_yaml,
                )
                workflow = workflow_renderer.render()
                return cls(content=workflow)
            except (YamlError, YamlKeyError) as ex:
                raise ex
            except Exception as ex:
                # Wrap all other "non-special" exceptions
                raise ValueError(f"Error rendering file: {file_path}") from ex

    def write_to_file(self, file_path: Path) -> None:
        logger.info("Write workflow file %s", file_path.name)
        file_path.write_text(self.content + "\n")


def _select_workflow_template(workflow_name: WorkflowChoice) -> Mapping[str, Path]:
    """
    Returns a mapping of workflow names to paths. Can be a single item or all workflow
    templates.
    """
    if workflow_name == ALL:
        return WORKFLOW_TEMPLATE_OPTIONS
    return {workflow_name: WORKFLOW_TEMPLATE_OPTIONS[workflow_name]}


def update_workflow(workflow_choice: WorkflowChoice, config: BaseConfig) -> None:
    """
    Updates a selected workflow or all workflows.
    """
    workflow_dict = _select_workflow_template(workflow_choice)
    logger.info(f"Selected workflow(s) to update: {list(workflow_dict.keys())}")

    workflow_patcher = None
    if config.github_workflow_patcher_yaml:
        workflow_patcher = WorkflowPatcher(
            github_template_dict=config.github_template_dict,
            file_path=config.github_workflow_patcher_yaml,
        )

    for workflow_name in workflow_dict:
        patch_yaml = None
        if workflow_patcher:
            patch_yaml = workflow_patcher.extract_by_workflow(
                workflow_name=workflow_name
            )

        try:
            workflow = Workflow.load_from_template(
                file_path=workflow_dict[workflow_name],
                github_template_dict=config.github_template_dict,
                patch_yaml=patch_yaml,
            )
            file_path = config.github_workflow_directory / f"{workflow_name}.yml"
            workflow.write_to_file(file_path=file_path)
        except YamlKeyError as ex:
            raise InvalidWorkflowPatcherEntryError(
                file_path=config.github_workflow_patcher_yaml, entry=ex.entry  # type: ignore
            ) from ex
